from abc import abstractmethod
from abc import ABC
import asyncio

from gql.transport.requests import RequestsHTTPTransport
from bergen.console import console
from bergen.wards.base import WardException
from gql.gql import gql

from gql.transport.aiohttp import AIOHTTPTransport
from bergen.schema import DataPoint
from bergen.query import  TypedGQL
from typing import TypeVar

T = TypeVar("T")




class MainWard(ABC):

    def __init__(self, client, loop=None):
        self.loop = loop or client.loop or asyncio.get_event_loop()
        self.host = client.config.host
        self.port = client.config.port
        self.protocol = "https" if client.config.secure else "http"
        self.token = client.auth.access_token

        assert self.token is not None, "Cannot create a Ward without acquiring a Token first"
        self._graphql_endpoint = f"{self.protocol}://{self.host}:{self.port}/graphql"
        self._headers = {"Authorization": f"Bearer {self.token}"}

        self.sync_transport = RequestsHTTPTransport(self._graphql_endpoint, headers=self._headers, verify=True, retries=3)
        self.sync_transport.connect()
        

    def run(self, the_query: TypedGQL, variables: dict = {}, **kwargs):
        query_node = gql(the_query.query)
        try:
            response = self.sync_transport.execute(query_node, variable_values=variables)
            if response.errors:
                raise WardException(f"Error: {self._graphql_endpoint} {str(response.errors)} called with Query {query_node} and variables {variables}")
            return the_query.extract(response.data)
        except:
            raise

    async def configure(self):
        self.async_transport = AIOHTTPTransport(url=self._graphql_endpoint, headers=self._headers)
        await self.async_transport.connect()


    async def run_async(self, the_query: TypedGQL, variables: dict = {}, **kwargs):
        query_node = gql(the_query.query)
        try:
            response = await self.async_transport.execute(query_node, variable_values=variables)

            if response.errors:
                raise WardException(f"Ward {self._graphql_endpoint}:" + str(response.errors))
            
            return the_query.extract(response.data)

        except:
            console.print_exception(show_locals=True)
            raise 
            
        
    async def disconnect(self):
        await self.async_transport.close()


    async def __aenter__(self):
        await self.configure()
        return self


    async def __aexit__(self, *args, **kwargs):
        await self.disconnect()




from bergen.enums import ClientType
from bergen.clients.default import Bergen


class ProviderBergen(Bergen):

    def __init__(self, *args, scopes= [],**kwargs) -> None:
        super().__init__(*args, client_type = ClientType.PROVIDER, scopes=scopes + ["host","provide"], **kwargs)

     



from bergen.queries.delayed.node import DETAIL_NODE_FR
from bergen.query import DelayedGQL


POD_QUERY = DelayedGQL("""
query Pod($id: ID){
  pod(id: $id){
    id
    status
    name
    unique
    channel
    template {
      id
      provider {
        name
      }
      node {"""

+ DETAIL_NODE_FR + 

"""
      }
    }
  }
}
""")



from abc import ABC
from bergen.query import TypedGQL
import requests
from bergen.schema import WardSettings

from bergen.wards.base import BaseWard, WardException


class BaseGraphQLWard(BaseWard, ABC):
    can_subscribe = False
    

    def __init__(self, client, settings: WardSettings, loop=None) -> None:
        super().__init__(client, settings, loop=loop)
        self._graphql_endpoint = f"{self.protocol}://{self.host}:{self.port}/graphql"
        self._headers = {"Authorization": f"Bearer {self.token}"}

        
    def run(self, the_query: TypedGQL, variables: dict = {}, **kwargs):
        query_node = the_query.query
        try:
            response = requests.post(self._graphql_endpoint, json={"query": query_node, "variables":variables}, headers=self._headers).json()
            if "errors" in response:
                raise WardException(f"Error: {self._graphql_endpoint} {str(response['errors'])} called with Query {query_node} and variables {variables}")
            return the_query.extract(response["data"])
        except:
            raise




        



from bergen.query import DelayedGQL


INPUTS_FR = """
  args {
    __typename
      key
      required
      description
      widget {
        __typename
        dependencies
        ... on QueryWidget {
          query  
        }
      }
      label
      ... on ModelArgPort {
        identifier
      }
      ... on IntArgPort {
        default
      }
  }
  kwargs {
    __typename
      key
      required
      description
      widget {
        __typename
        dependencies
        ... on QueryWidget {
          query  
        }
      }
      label
      ... on ModelKwargPort {
        identifier
      }
      ... on IntKwargPort {
        default
      }
  }
"""


OUTPUTS_FR = """
  returns {
    __typename
      key
      description
      ... on ModelReturnPort {
        identifier
      }
  }
"""



DETAIL_NODE_FR = """
  name
  package
  interface
  id
  type
""" + INPUTS_FR + OUTPUTS_FR 


NODE_QUERY = DelayedGQL("""
query Node($id: ID, $package: String, $interface: String){
  node(id: $id, package: $package, interface: $interface){
    """+ DETAIL_NODE_FR +"""
  }
}
""")


NODE_FILTER_QUERY = DelayedGQL("""
query NodeFilter($name: String){
  nodes(name: $name){
    id
    name
    repository {
      name
    }
    description
  }
}
""")


CREATE_NODE_MUTATION = DelayedGQL("""
  mutation CreateNodeMutation(
    $description: String!,
    $args: [ArgPortInput]!,
    $kwargs: [KwargPortInput]!,
    $returns: [ReturnPortInput]!,
    $package: String!, $interface: String!,
    $name: String!
    $type: NodeTypeInput){
  createNode(description: $description, args: $args, kwargs: $kwargs, returns: $returns, package:$package, interface: $interface, name: $name, type: $type){
    """+ DETAIL_NODE_FR +"""
  }
}
""")


UPDATE_OR_CREATE_NODE = DelayedGQL("""
  mutation CreateNodeMutation($description: String!, $args: [ArgPortInput]!, $kwargs: [KwargPortInput]!, $returns: [ReturnPortInput]!, $interface: String!, $name: String!, $type: NodeTypeInput){
  createNode(description: $description, args: $args, kwargs: $kwargs, returns: $returns, interface: $interface, name: $name, type: $type){
    """+ DETAIL_NODE_FR +"""
  }
}
""")


from abc import abstractmethod
from abc import ABC
import asyncio
import aiohttp
import requests
from bergen.console import console
from bergen.wards.base import WardException
from bergen.query import  TypedGQL
from typing import TypeVar

T = TypeVar("T")


class MainWard(ABC):

    def __init__(self, client, loop=None):
        self.loop = loop or client.loop or asyncio.get_event_loop()
        self.host = client.config.host
        self.port = client.config.port
        self.protocol = "https" if client.config.secure else "http"
        self.token = client.auth.access_token

        assert self.token is not None, "Cannot create a Ward without acquiring a Token first"
        self._graphql_endpoint = f"{self.protocol}://{self.host}:{self.port}/graphql"
        self._headers = {"Authorization": f"Bearer {self.token}"}
        

    def run(self, the_query: TypedGQL, variables: dict = {}, **kwargs):
        query_node = the_query.query
        try:
            response = requests.post(self._graphql_endpoint, json={"query": query_node, "variables":variables}, headers=self._headers).json()
            if "errors" in response:
                raise WardException(f"Error: {self._graphql_endpoint} {str(response['errors'])} called with Query {query_node} and variables {variables}")
            return the_query.extract(response["data"])
        except:
            raise

    async def configure(self):
        self.async_session = aiohttp.ClientSession(headers=self._headers)


    async def run_async(self, the_query: TypedGQL, variables: dict = {}, **kwargs):
        query_node = the_query.query
        try:
            async with self.async_session.post(self._graphql_endpoint, json={"query": query_node, "variables": variables}) as resp:
                result = await resp.json() 

                if "errors" in result:
                    raise  WardException(f"Ward {self._graphql_endpoint}:" + str(result["errors"]))

                return the_query.extract(result["data"])

        except:
            console.print_exception(show_locals=True)
            raise 
            
        
    async def disconnect(self):
        await self.async_session.close()


    async def __aenter__(self):
        await self.configure()
        return self


    async def __aexit__(self, *args, **kwargs):
        await self.disconnect()




from bergen.messages import *

def build_assign_message(reference, reservation, args, kwargs, with_progress=False, bounced=None, persist=False):
    assert reference is not None, "Must have a reference"

    data = {
                                    "reservation": reservation,
                                    "args": args, 
                                    "kwargs": kwargs,
    }

    meta = {
                                    "reference": reference, 
                                    "extensions": {
                                        "with_progress": with_progress,
                                        "persist": persist
                                    }
    }

    if bounced:
        meta = {**meta, "token": bounced}
        return BouncedAssignMessage(data=data, meta=meta)

    else:
        return AssignMessage(data=data, meta=meta)



def build_unassign_messsage(reference, assignation, with_progress=False, bounced=None):
    assert reference is not None, "Must have a reference"
    data= {
                                        "assignation": assignation
    }
                                    
    meta={
                                    "reference": reference, 
                                    "extensions": {
                                        "with_progress": with_progress
                                    }
    }

    if bounced:
        meta = {**meta, "token": bounced}
        return BouncedUnassignMessage(data=data, meta=meta)

    else:
        return UnassignMessage(data=data, meta=meta)


def build_reserve_message(reference, node_id: str, template_id: str, provision: str, params_dict: dict = {}, with_progress=False, bounced=None):
    assert reference is not None, "Must have a reference"
    assert node_id is not None or template_id is not None, "Please provide either a node_id or template_id"

    data={
                                "node": node_id, 
                                "template": template_id, 
                                "provision": provision,
                                "params": params_dict,
    }
    meta={
        "reference": reference,
        "extensions": {
            "with_progress": with_progress,
        }
    }

    if bounced:
        meta = {**meta, "token": bounced}
        return BouncedReserveMessage(data=data, meta=meta)

    else:
        return ReserveMessage(data=data, meta=meta)



def build_unreserve_messsage(reference, reservation, with_progress=False, bounced=None):
    assert reference is not None, "Must have a reference"
    data= {
                                        "reservation": reservation
    }
    
    meta={
                                    "reference": reference, 
                                    "extensions": {
                                        "with_progress": with_progress
                                    }
    }

    if bounced:
        meta = {**meta, "token": bounced}
        return BouncedUnreserveMessage(data=data, meta=meta)

    else:
        return UnreserveMessage(data=data, meta=meta)




from abc import ABC

import aiohttp
from bergen.wards.bare.base import BaseGraphQLWard
from bergen.wards.base import WardException
from bergen.query import TypedGQL
import logging
from bergen.console import console


class AIOHttpWard(BaseGraphQLWard):
    can_subscribe = False

    async def connect(self):
        self.async_session = aiohttp.ClientSession(headers=self._headers)

    async def negotiate(self):
        if self.needs_negotiation:
            query_node = """
                mutation Negotiate {
                    negotiate
                }
            """
            async with self.async_session.post(self._graphql_endpoint, json={"query": query_node}) as resp:
                result = await resp.json() 
                return result["data"]["negotiate"]
        else:
            return None
    
    async def run_async(self, the_query: TypedGQL, variables: dict = {}, **kwargs):
        query_node = the_query.query
        try:
            async with self.async_session.post(self._graphql_endpoint, json={"query": query_node, "variables": variables}) as resp:
                result = await resp.json() 

                if "errors" in result:
                    raise  WardException(f"Ward {self._graphql_endpoint}:" + str(result["errors"]))

                return the_query.extract(result["data"])

        except:
            console.print_exception(show_locals=True)
            raise 
            
        
    async def disconnect(self):
        await self.async_session.close()


from .provide_critical import ProvideCriticalMessage
from .provide import ProvideMessage
from .provide_done import ProvideDoneMessage
from .provide_log import ProvideLogMessage
from .bounced_provide import BouncedProvideMessage
from .provide_transition import ProvideTransitionMessage


from bergen.schema import DataModel
from bergen.query import TypedGQL


HOST_GQL = TypedGQL("""
    mutation Host($identifier: String!, $extenders: [String]){
        host(identifier: $identifier, extenders: $extenders){
            id
            identifier
            extenders
            point {
                inward
            }
        }
    }
""", DataModel)


import asyncio
from bergen.debugging import DebugLevel
from bergen.handlers import AssignHandler, ReserveHandler, ProvideHandler
import contextvars

loop_context = contextvars.ContextVar("loop_context", default=None)
assign_handler_context  = contextvars.ContextVar('assign_handler', default=None)
provide_handler_context = contextvars.ContextVar("provide_handler", default=None)


async def log_async(message, level: DebugLevel = DebugLevel.INFO):
    handler: AssignHandler = assign_handler_context.get()
    await handler.log(message, level)


def log(message: str, level: DebugLevel = DebugLevel.INFO):
    """ Logs a message

    Depending on both the configuration of Arkitekt and the overwrite set on the
    Assignment, this logging will be sent (and persisted) on the Arkitekt server

    Args:
        message (sr): The Message you want to send
        level (DebugLevel, optional): The level of the log. Defaults to DebugLevel.INFO.

    Returns:
        [Future]: Returns a future if currently running in an event loop
    """
    try:
        event_loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = loop_context.get()
        loop.create_task(log_async(message, level))
    else:
        if event_loop.is_running():
            return log_async(message, level=level)


from bergen.console import console

GraphQLWard = None
MainWard = None

try:
    from bergen.wards.gql.aiohttp import AIOHttpGraphQLWard
    GraphQLWard = AIOHttpGraphQLWard
except ImportError as e:
    console.log("No installed GQL Transport. Using bare aiohttp")
    from bergen.wards.bare.aiohttp import AIOHttpWard
    GraphQLWard = AIOHttpWard


try:
    from bergen.wards.gql.main import MainWard as GQLMainWard
    MainWard = GQLMainWard

except ImportError as e:
    console.log("No installed GQL Transport. Using bare aiohttp")
    from bergen.wards.bare.main  import MainWard as BareMainWard
    MainWard = BareMainWard


import logging
from abc import ABC

from bergen.console import console
from bergen.query import GQL, TypedGQL
from bergen.wards.base import BaseWard
from bergen.wards.gql.base import BaseGraphQLWard, GraphQLException
from gql.gql import gql
from gql.transport.aiohttp import AIOHTTPTransport
from gql.transport.aiohttp import log as aiohttp_logger
from gql.transport.requests import RequestsHTTPTransport

aiohttp_logger.setLevel(logging.WARNING)


class AIOHttpGraphQLWard(BaseGraphQLWard):
    can_subscribe = False

    async def connect(self):
        self.async_transport = AIOHTTPTransport(url=self._graphql_endpoint, headers=self._headers)
        await self.async_transport.connect()

    async def negotiate(self):
        if self.needs_negotiation:
            query_node = gql("""
                mutation Negotiate {
                    negotiate
                }
            """)
            response = await self.async_transport.execute(query_node, variable_values={})
            return response.data["negotiate"]
        else:
            return None
    

    async def run_async(self, the_query: TypedGQL, variables: dict = {}, **kwargs):
        query_node = gql(the_query.query)
        try:
            response = await self.async_transport.execute(query_node, variable_values=variables)

            if response.errors:
                raise GraphQLException(f"Ward {self._graphql_endpoint}:" + str(response.errors))
            
            return the_query.extract(response.data)

        except:
            console.print_exception(show_locals=True)
            raise 
            
        
    async def disconnect(self):
        await self.async_transport.close()
